import torch
from torch import nn
from typing import Union, List
from ._text_cnn import TextCNN
from ._attention import RNNAttention, CNNRNNAttention, RNNCNNAttention, ResRNNCNNAttention
from ._embedding import EmbeddingLayer


class EmbeddingClassifier(nn.Module):
	"""
	Examples
	--------
	>>> from nlpx.text_token import TokenEmbedding
	>>> from nlpx.models import TextCNN, RNNAttention, EmbeddingClassifier
	>>> tokenizer = TokenEmbedding(pretrained_path)
	>>> attn = RNNAttention(tokenizer.embed_dim, num_heads=2, out_features=len(classes))
	>>> classifier = EmbeddingClassifier(atten, embedding=tokenizer.embedding)
	>>> classifier = EmbeddingClassifier(atten, vocab_size=tokenizer.vocab_size, embed_dim=tokenizer.embed_dim)
	"""

	def __init__(self, classifier, embedding: Union[nn.Embedding, torch.Tensor, List] = None, vocab_size: int = None,
	             embed_dim: int = None):
		"""
		:param classifier: 分类器
		:param embedding: 输入的embedding，可以是nn.Embedding，torch.Tensor，list
		:param vocab_size: vocab size
		:param embed_dim: word embedding维度
		"""
		super().__init__()
		self.classifier = classifier
		self.embedding = EmbeddingLayer(embed_dim, vocab_size, embedding)

	def forward(self, input_ids: torch.Tensor, labels: torch.LongTensor = None):
		embedding = self.embedding(input_ids)
		return self.classifier(embedding, labels)


class TextCNNClassifier(EmbeddingClassifier):
	"""
	Examples
	--------
	>>> from nlpx.text_token import Tokenizer
	>>> from nlpx.models import TextCNNClassifier
	>>> tokenizer = Tokenizer(corpus)
	>>> classifier = TextCNNClassifier(embed_dim, len(classes), vocab_size=tokenizer.vocab_size)
	"""

	def __init__(self, embed_dim: int, num_classes: int, embedding: Union[nn.Embedding, torch.Tensor, List] = None,
	             vocab_size: int = None, kernel_sizes=(2, 3, 4), cnn_channels: int = 64, activation=nn.ReLU(inplace=True),
				 num_hidden_layer: int = 0, batch_norm=False, layer_norm=False, drop_out: float = 0.0):
		classifier = TextCNN(embed_dim, kernel_sizes, cnn_channels, num_classes, activation, num_hidden_layer, batch_norm,
						   layer_norm, drop_out)
		super().__init__(classifier, embedding, vocab_size, embed_dim)


class RNNAttentionClassifier(EmbeddingClassifier):
	"""
	Examples
	--------
	>>> from nlpx.text_token import Tokenizer
	>>> from nlpx.models import RNNAttentionClassifier
	>>> tokenizer = Tokenizer(corpus)
	>>> classifier = RNNAttentionClassifier(embed_dim, len(classes), vocab_size=tokenizer.vocab_size)
	"""

	def __init__(self, embed_dim: int, num_classes: int, embedding: Union[nn.Embedding, torch.Tensor, List] = None,
	             vocab_size: int = None, hidden_size: int = 64, num_layers: int = 1, num_heads: int = 1, rnn=nn.GRU,
	             drop_out: float = 0.0):
		classifier = RNNAttention(embed_dim, hidden_size, num_layers, num_heads, num_classes, rnn, drop_out)
		super().__init__(classifier, embedding, vocab_size, embed_dim)


class CNNRNNAttentionClassifier(EmbeddingClassifier):
	"""
	Examples
	--------
	>>> from nlpx.text_token import Tokenizer
	>>> from nlpx.models import CNNRNNAttentionClassifier
	>>> tokenizer = Tokenizer(corpus)
	>>> classifier = CNNRNNAttentionClassifier(embed_dim, num_classes=len(classes), vocab_size=tokenizer.vocab_size)
	"""

	def __init__(self, embed_dim: int, seq_length: int = 16, cnn_channels: int = None, kernel_sizes=(2, 3, 4),
	             activation=nn.ReLU(inplace=True), num_classes: int = 2, bias=False,
	             embedding: Union[nn.Embedding, torch.Tensor, List] = None, vocab_size: int = None,
				 hidden_size: int = 64, num_layers: int = 1,num_heads: int = 2, rnn=nn.GRU, drop_out: float = 0.0):
		"""
		:param embed_dim: word embedding维度
		:param seq_length: 句子序列长度
		:param cnn_channels: CNN out_channels
		:param kernel_sizes: size of each CNN kernel
		:param activation: CNN 激活函数
		:param num_classes: 类别数
		:param bias:
		:param embedding: 出入的embedding
		:param vocab_size: vocab size
		:param hidden_size: RNN的hidden_size, RNN隐藏层维度
		:param num_layers: RNN的num_layers, RNN层数
		:param num_heads: 抽头数
		:param rnn: 所用的RNN模型：GRU和LSTM，默认是GRU
		:param drop_out：
		"""
		cnn_channels = cnn_channels or embed_dim
		classifier = CNNRNNAttention(embed_dim, seq_length, cnn_channels, kernel_sizes, activation, hidden_size,
		                             num_layers, num_heads, num_classes, rnn, drop_out, bias)
		super().__init__(classifier, embedding, vocab_size, embed_dim)


class RNNCNNAttentionClassifier(EmbeddingClassifier):
	"""
	Examples
	--------
	>>> from nlpx.text_token import Tokenizer
	>>> from nlpx.models import CNNRNNAttentionClassifier
	>>> tokenizer = Tokenizer(corpus)
	>>> classifier = RNNCNNAttentionClassifier(embed_dim, num_classes=len(classes), vocab_size=tokenizer.vocab_size)
	"""

	def __init__(self, embed_dim: int, seq_length: int = 16, cnn_channels: int = None, kernel_sizes=(2, 3, 4),
	             activation=nn.ReLU(inplace=True), num_classes: int = 2, bias=False,
	             embedding: Union[nn.Embedding, torch.Tensor, List] = None, vocab_size: int = None,
				 hidden_size: int = 64, num_layers: int = 1,num_heads: int = 2, rnn=nn.GRU, drop_out: float = 0.0):
		"""
		:param embed_dim: word embedding维度
		:param seq_length: 句子序列长度
		:param cnn_channels: CNN out_channels
		:param kernel_sizes: size of each CNN kernel
		:param activation: CNN 激活函数
		:param num_classes: 类别数
		:param bias:
		:param embedding: 出入的embedding
		:param vocab_size: vocab size
		:param hidden_size: RNN的hidden_size, RNN隐藏层维度
		:param num_layers: RNN的num_layers, RNN层数
		:param num_heads: 抽头数
		:param rnn: 所用的RNN模型：GRU和LSTM，默认是GRU
		:param drop_out：
		"""
		cnn_channels = cnn_channels or embed_dim
		classifier = RNNCNNAttention(embed_dim, seq_length, cnn_channels, kernel_sizes, activation, hidden_size,
		                             num_layers, num_heads, num_classes, rnn, drop_out, bias)
		super().__init__(classifier, embedding, vocab_size, embed_dim)


class ResRNNCNNAttentionClassifier(EmbeddingClassifier):
	"""
	Examples
	--------
	>>> from nlpx.text_token import Tokenizer
	>>> from nlpx.models import CNNRNNAttentionClassifier
	>>> tokenizer = Tokenizer(corpus)
	>>> classifier = ResRNNCNNAttentionClassifier(embed_dim, num_classes=len(classes), vocab_size=tokenizer.vocab_size)
	"""

	def __init__(self, embed_dim: int, seq_length: int = 16, cnn_channels: int = None, kernel_sizes=(2, 3, 4),
	             activation=nn.ReLU(inplace=True), num_classes: int = 2, bias=False,
	             embedding: Union[nn.Embedding, torch.Tensor, List] = None, vocab_size: int = None,
				 hidden_size: int = 64, num_layers: int = 1,num_heads: int = 2, rnn=nn.GRU, batch_norm=False,
				 drop_out: float = 0.0):
		"""
		:param embed_dim: word embedding维度
		:param seq_length: 句子序列长度
		:param cnn_channels: CNN out_channels
		:param kernel_sizes: size of each CNN kernel
		:param activation: CNN 激活函数
		:param num_classes: 类别数
		:param bias:
		:param embedding: 出入的embedding
		:param vocab_size: vocab size
		:param hidden_size: RNN的hidden_size, RNN隐藏层维度
		:param num_layers: RNN的num_layers, RNN层数
		:param num_heads: 抽头数
		:param rnn: 所用的RNN模型：GRU和LSTM，默认是GRU
		:param batch_norm：
		:param drop_out：
		"""
		cnn_channels = cnn_channels or embed_dim
		classifier = ResRNNCNNAttention(embed_dim, seq_length, cnn_channels, kernel_sizes, activation, hidden_size,
		                             num_layers, num_heads, num_classes, rnn, drop_out, batch_norm, bias)
		super().__init__(classifier, embedding, vocab_size, embed_dim)
	



import torch
from torch import nn


class CNNLayer(nn.Module):
	
	def __init__(self, embed_dim: int, seq_length: int = 16, out_channels: int = None, kernel_sizes=(2, 3, 4),
	             activation=nn.ReLU(inplace=True), batch_norm=False, bias=False):
		"""
		:param embed_dim: word embedding维度
		:param seq_length: 句子序列长度
		:param out_channels: CNN out_channels, default embed_dim
		:param kernel_sizes: size of each CNN kernel
		:param activation: CNN 激活函数
		:param batch_norm: 是否批正则化
		:param bias:
		"""
		super().__init__()
		self.batch_norm = batch_norm
		out_channels = out_channels or embed_dim
		self.convs = nn.ModuleList([
			nn.Sequential(
				nn.Conv1d(in_channels=embed_dim, out_channels=out_channels, kernel_size=kernel_size, bias=bias),
				activation,  # inplace为True，将会改变输入的数据 ，否则不会改变原输入，只会产生新的输出
				nn.AdaptiveMaxPool1d(seq_length)
			) for kernel_size in kernel_sizes
		])
		if batch_norm:
			self.norm = nn.BatchNorm1d(num_features=out_channels)
	
	def forward(self, inputs: torch.Tensor):
		"""
		:param inputs: [(batch_size, sequence_length, embed_dim)]
		:return: [(batch_size, seq_length * len(kernel_sizes), out_channels)]
		"""
		inputs = inputs.transpose(2, 1)
		output = torch.cat([conv(inputs) for conv in self.convs], dim=-1)
		if self.batch_norm:
			output = self.norm(output)
		return output.transpose(2, 1)



import torch
import numpy as np
import pandas as pd
from torch import optim
from pathlib import Path
from typing import Union, List
from torch.nn import functional as F
from torch.utils.data import DataLoader, TensorDataset, Dataset

from nlpx.llm import TokenizeVec
from nlpx.text_token import TokenEmbedding
from nlpx.training import Trainer, SimpleTrainer, evaluate


class ModelWrapper:

	def __init__(self, model_path: Union[str, Path] = None, classes: List[str] = None,
					device=torch.device('cuda:0' if torch.cuda.is_available() else 'cpu')):
		self.classes = classes
		self.device = device
		self.model = torch.load(model_path, map_location=device) if model_path else None

	def train(self, model, train_set: Dataset, eval_set: Dataset, collate_fn=None, max_iter=100,
				optimizer=optim.AdamW, learning_rate=0.001, T_max: int = 0,
				batch_size=8, eval_batch_size=16,
				num_workers=0, num_eval_workers=0,
                pin_memory: bool = False, pin_memory_device: str = "",
                persistent_workers: bool = False,
				early_stopping_rounds: int = 10,
	            print_per_rounds: int = 1):
		trainer = Trainer(model, train_set, eval_set, collate_fn,
						  max_iter, optimizer, learning_rate, T_max,
						  batch_size, eval_batch_size,
						  num_workers, num_eval_workers,
						  pin_memory, pin_memory_device,
		                  persistent_workers,
						  early_stopping_rounds,  # 早停，等10轮决策，评价指标不在变化，停止
						  print_per_rounds,
						  self.device)
		self.model = trainer.train()

	def predict(self, X: torch.Tensor):
		logits = self.logits(X)
		return logits.argmax(1)

	def predict_classes(self, X: torch.Tensor):
		assert self.classes is not None, 'classes must be specified'
		pred = self.predict(X)
		return [self.classes[i] for i in pred.detach().numpy().ravel()]

	def predict_proba(self, X: torch.Tensor):
		logits = self.logits(X)
		result = F.softmax(logits, dim=1).max(1)
		return result.indices, result.values

	def predict_classes_proba(self, X: torch.Tensor):
		assert self.classes is not None, 'classes must be specified'
		indices, values = self.predict_proba(X)
		return [self.classes[i] for i in indices.detach().numpy().ravel()], values

	def logits(self, X: torch.Tensor):
		self.model.eval()
		with torch.no_grad():
			logits = self.model(X)
		return logits

	def evaluate(self, eval_set: Dataset, batch_size=16, num_workers=0, max_length: int = 0,  collate_fn=None):
		eval_loader = DataLoader(dataset=eval_set, batch_size=batch_size,
								 num_workers=num_workers, collate_fn=collate_fn)
		_, acc = evaluate(self.model, eval_loader, self.device)
		return acc

	def save(self, model_path: Union[str, Path] = './best_model.pt'):
		torch.save(model_path, self.model)

	def load(self, model_path: Union[str, Path] = './best_model.pt'):
		self.model = torch.load(model_path, map_location=self.device)


class SimpleModelWrapper(ModelWrapper):

	def __init__(self, tokenize_vec: Union[TokenizeVec, TokenEmbedding], model_path: Union[str, Path] = None,
				 classes: List[str] = None, device=torch.device('cuda:0' if torch.cuda.is_available() else 'cpu')):
		super().__init__(model_path, classes, device)
		self.tokenize_vec = tokenize_vec

	def train(self, model, texts: Union[List[str], np.ndarray, pd.Series], y: Union[torch.LongTensor, List, np.ndarray],
			  max_length: int = 0, eval_size=0.2, random_state=None, collate_fn=None,max_iter=100,
			  optimizer=optim.AdamW, learning_rate=0.001, T_max: int = 0,
			  batch_size=8, eval_batch_size=16,
			  num_workers=0, num_eval_workers=0,
              pin_memory: bool = False, pin_memory_device: str = "",
              persistent_workers: bool = False,
			  early_stopping_rounds: int = 10,
	          print_per_rounds: int = 1, n_jobs=-1, cut_fn=None):
		X = self.get_vec(texts, max_length=max_length, n_jobs=n_jobs)
		if isinstance(y, List) or isinstance(y, np.ndarray):
			y = torch.tensor(y, dtype=torch.long)
		trainer = SimpleTrainer(model, X, y, eval_size, random_state, collate_fn,
								 max_iter, optimizer, learning_rate, T_max,
								 batch_size, eval_batch_size,
								 num_workers, num_eval_workers,
							     pin_memory, pin_memory_device,
			                     persistent_workers,
								 early_stopping_rounds,
								 print_per_rounds,
								 self.device)
		self.model = trainer.train()

	def predict(self, texts: List[str], max_length: int = 0, n_jobs=-1, cut_fn=None):
		logits = self.logits(texts, max_length, n_jobs=n_jobs, cut_fn=cut_fn)
		return logits.argmax(1)

	def predict_classes(self, texts: List[str], max_length: int = 0, n_jobs=-1, cut_fn=None):
		assert self.classes is not None, 'classes must be specified'
		pred = self.predict(texts, max_length, n_jobs=n_jobs, cut_fn=cut_fn)
		return [self.classes[i] for i in pred.detach().numpy().ravel()]
	
	def predict_proba(self, texts: List[str], max_length: int = 0, n_jobs=-1, cut_fn=None):
		logits = self.logits(texts, max_length, n_jobs=n_jobs, cut_fn=cut_fn)
		result = F.softmax(logits, dim=1).max(1)
		return result.indices, result.values

	def predict_classes_proba(self, texts: List[str], max_length: int = 0, n_jobs=-1, cut_fn=None):
		assert self.classes is not None, 'classes must be specified'
		indices, values = self.predict_proba(texts, max_length, n_jobs, cut_fn=cut_fn)
		return [self.classes[i] for i in indices.detach().numpy().ravel()], values

	def logits(self, texts: List[str], max_length: int = 0, n_jobs=-1, cut_fn=None):
		X = self.get_vec(texts, max_length, n_jobs=n_jobs, cut_fn=cut_fn)
		return super().logits(X)

	def evaluate(self, texts: Union[str, List[str], np.ndarray, pd.Series], y: Union[torch.LongTensor, List, np.ndarray], batch_size=16, num_workers=0,
				 max_length: int = 0, collate_fn=None, n_jobs=-1, cut_fn=None):
		X = self.get_vec(texts, max_length, n_jobs=n_jobs, cut_fn=cut_fn)
		if isinstance(y, List) or isinstance(y, np.ndarray):
			y = torch.tensor(y, dtype=torch.long)
		eval_set = TensorDataset(X, y)
		return super().evaluate(eval_set, batch_size=batch_size, num_workers=num_workers, collate_fn=collate_fn)

	def get_vec(self, texts: Union[str, List[str], np.ndarray, pd.Series], max_length: int, n_jobs: int, cut_fn=None):
		if isinstance(texts, str):
			texts = [texts]

		if isinstance(self.tokenize_vec, TokenizeVec):
			return self.tokenize_vec.parallel_encode_plus(texts, max_length=max_length, padding='max_length',
															truncation=True, add_special_tokens=True,
															return_token_type_ids=True,return_attention_mask=True,
															return_tensors='pt', n_jobs=n_jobs)
		elif isinstance(self.tokenize_vec, TokenEmbedding):
			return self.tokenize_vec(texts, max_length, cut_fn=cut_fn)

		raise ValueError("Invalid tokenize_vec, it must be a TokenizeVec or TokenEmbedding.")



import torch
from torch import nn


class RNNLayer(nn.Module):
	"""
	Examples
	--------
	>>> from nlpx.text_token import Tokenizer
	>>> from nlpx.models import RNNAttention
	>>> tokenizer = Tokenizer(corpus)
	>>> classifier = RNNLayer(embed_dim)
	"""
	
	def __init__(self, embed_dim: int, hidden_size: int = 64, num_layers: int = 1, rnn=nn.GRU, drop_out: float = 0.0,
	             batch_norm=False):
		"""
		:param embed_dim: RNN的input_size，word embedding维度
		:param hidden_size: RNN的hidden_size, RNN隐藏层维度
		:param num_layers: RNN的num_layers, RNN层数
		:param num_layers: RNN的num_layers, RNN层数
		:param rnn: 所用的RNN模型：GRU和LSTM，默认是GRU
		:param drop_out：
		:param batch_norm：是否批正则化
		"""
		
		super().__init__()
		self.batch_norm = batch_norm
		self.rnn = rnn(input_size=embed_dim, hidden_size=hidden_size, num_layers=num_layers, bidirectional=True,
		               batch_first=True, dropout=drop_out)
		if batch_norm:
			self.norm = nn.BatchNorm1d(num_features=(hidden_size << 1))
			
	def forward(self, inputs: torch.Tensor):
		"""
		:param inputs: [(batch_size, sequence_length, embed_dim)]
		:return: [(batch_size, sequence_length, 2 * hidden_size)]
		"""
		output, _ = self.rnn(inputs)  # [(batch_size, sequence_length, 2 * hidden_size)]
		if self.batch_norm:
			output = self.norm(output.transpose(2, 1)).transpose(2, 1)
		return output



import torch
from torch import nn
from ._cnn import CNNLayer


class EmbeddingLayer(nn.Module):
	
	def __init__(self, embed_dim: int, vocab_size: int = None, embedding=None):
		"""
		:param embed_dim: word embedding维度
		:param vocab_size: vocab size
		:param embedding: 输入的embedding，可以是nn.Embedding，torch.Tensor，list
		"""
		super().__init__()
		if embedding is None:
			assert vocab_size, 'vocab_size must be ge 0'
			self.embedding = nn.Embedding(vocab_size, embed_dim)
		elif isinstance(embedding, nn.Embedding):
			self.embedding = embedding
		elif isinstance(embedding, torch.Tensor):
			self.embedding = nn.Embedding.from_pretrained(embedding)
		else:
			self.embedding = nn.Embedding.from_pretrained(torch.tensor(embedding, dtype=torch.float))
			
	def forward(self, input_ids: torch.LongTensor):
		"""
		:param input_ids: [(batch_size, sequence_length, embed_dim)]
		:return: [(batch_size, sequence_length, embed_dim)]
		"""
		return self.embedding(input_ids)


class CNNEmbedding(nn.Module):
	
	def __init__(self, embed_dim: int, seq_length: int = 16, out_channels: int = None, kernel_sizes=(2, 3, 4),
	             activation=nn.ReLU(inplace=True), embedding=None, vocab_size: int = None, batch_norm=False, bias=False):
		"""
		:param embed_dim: word embedding维度
		:param seq_length: 句子序列长度
		:param out_channels: CNN out_channels, default embed_dim
		:param kernel_sizes: size of each CNN kernel
		:param activation: CNN 激活函数
		:param embedding: 输入的embedding，可以是nn.Embedding，torch.Tensor，list
		:param vocab_size: vocab size
		:param batch_norm: 是否批正则化
		:param bias:
		"""
		super().__init__()
		self.embedding = EmbeddingLayer(embed_dim, vocab_size, embedding)
		out_channels = out_channels or embed_dim
		self.cnn = CNNLayer(embed_dim, seq_length, out_channels, kernel_sizes, activation, batch_norm, bias)
	
	def forward(self, input_ids: torch.LongTensor):
		"""
		:param input_ids: [(batch_size, sequence_length, embed_dim)]
		:return: [(batch_size, seq_length * len(kernel_sizes), out_channels)]
		"""
		embeddings = self.embedding(input_ids)
		return self.cnn(embeddings)



import torch
from torch import nn
from torch.nn import functional as F


class TextCNN(nn.Module):
    """
    Examples
    --------
    >>> import torch
    >>> from nlpx.models import TextCNN
    >>> X = torch.randn(batch_size, 10, word_dim)
    >>> targets = torch.randint(0, num_classes, (batch_size,))
    >>> model = TextCNN(embed_dim, cnn_channels=64, kernel_sizes=(2, 3, 4), out_features=num_classes)
    >>> output = model(X)
    >>> loss, output = model(X, targets)
    """

    def __init__(self, embed_dim: int, kernel_sizes=(2, 3, 4), cnn_channels: int = 64, out_features: int = 2,
                 activation=nn.ReLU(inplace=True), num_hidden_layer: int = 0, batch_norm=False, layer_norm=False,
                 drop_out: float = 0.0):
        """ TextCNN model
        Parameters
        ----------
        embed_dim: int, dim of embedding, in_channels of cnn
        cnn_channels: int, out_channels of cnn
        kernel_sizes: size of each cnn kernel
        out_features: dim of output
        activation:
        num_hidden_layer:
        drop_out:
        """

        super().__init__()
        if batch_norm:
            self.convs = nn.ModuleList([
                nn.Sequential(
                    nn.Conv1d(in_channels=embed_dim, out_channels=cnn_channels, kernel_size=kernel_size, bias=False),
                    nn.BatchNorm1d(num_features=cnn_channels),
                    activation,  # inplace为True，将会改变输入的数据 ，否则不会改变原输入，只会产生新的输出
                    nn.AdaptiveMaxPool1d(1)
                ) for kernel_size in kernel_sizes
            ])
        else:
            self.convs = nn.ModuleList([
                nn.Sequential(
                    nn.Conv1d(in_channels=embed_dim, out_channels=cnn_channels, kernel_size=kernel_size, bias=False),
                    activation,  # inplace为True，将会改变输入的数据 ，否则不会改变原输入，只会产生新的输出
                    nn.AdaptiveMaxPool1d(1)
                ) for kernel_size in kernel_sizes
            ])

        self.num_hidden_layer = num_hidden_layer
        num_features = cnn_channels * len(kernel_sizes)
        if num_hidden_layer and num_hidden_layer > 0:
            if layer_norm:
                self.hidden_layers = nn.ModuleList([
                    nn.Sequential(
                        nn.Linear(in_features=num_features, out_features=num_features),
                        nn.LayerNorm(normalized_shape=num_features),
                        activation
                    ) for _ in range(num_hidden_layer)
                ])
            else:
                self.hidden_layers = nn.ModuleList([
                    nn.Sequential(
                        nn.Linear(in_features=num_features, out_features=num_features),
                        activation
                    ) for _ in range(num_hidden_layer)
                ])

        self.fc = nn.Linear(in_features=num_features, out_features=out_features)
        if 0.0 < drop_out < 1.0:
            self.fc = nn.Sequential(
                nn.Dropout(drop_out),
                self.fc
            )

    def forward(self, inputs: torch.Tensor, labels: torch.LongTensor = None):
        """
        :param inputs: [(batch_size, sequence_length, embed_dim)]
        :param labels: [long]
        """

        # Conv1d期望输入的形状是：(batch_size, embed_dim, sequence_length)
        input_embeddings = inputs.transpose(2, 1)

        # conv输出的形状是：(batch_size, cnn_channels, 1)
        # output的形状是：(batch_size, cnn_channels * len(kernel_sizes), 1)
        output = torch.cat([conv(input_embeddings) for conv in self.convs], dim=1)

        # 确保张量是连续的
        output = output.contiguous()
        # output：(batch_size, cnn_channels * len(kernel_sizes))
        output = output.view(inputs.size(0), -1)

        if self.num_hidden_layer and self.num_hidden_layer > 0:
            for hidden_layer in self.hidden_layers:
                output = hidden_layer(output)

        logits = self.fc(output)

        if labels is None:
            return logits

        loss = F.cross_entropy(logits, labels)
        return loss, logits



import torch
import numpy as np
import pandas as pd
from typing import Union, List
from torch.utils.data import Dataset

from nlpx.llm import TokenizeVec
from nlpx.text_token import BaseTokenizer, TokenEmbedding
from nlpx.text_token.utils import get_texts_max_length


class TokenDataset(Dataset):
	""" Token的长度不一样
	返回的是(token, label) 不是Tensor, 必须经过PaddingTokenCollator
	"""

	def __init__(self, tokens: Union[List[int], np.ndarray],
				 labels: Union[List, np.ndarray, pd.Series]):
		super().__init__()
		self.tokens = tokens
		self.labels = labels.values if isinstance(labels, pd.Series) else labels

	def __getitem__(self, index: int):
		return self.tokens[index], self.labels[index]

	def __len__(self):
		return len(self.labels)


class SameLengthTokenDataset(Dataset):
	""" Token已经truncate和padding, 长度一样
	返回的是Tensor, 不需要经过collate_fn
	"""

	def __init__(self, tokens: Union[List[int], np.ndarray, torch.LongTensor],
				 labels: Union[List, np.ndarray, pd.Series, torch.LongTensor]):
		super().__init__()
		self.tokens = tokens if isinstance(tokens, torch.LongTensor) else torch.tensor(tokens, dtype=torch.long)
		labels = labels.values if isinstance(labels, pd.Series) else labels
		self.labels = labels if isinstance(labels, torch.LongTensor) else torch.tensor(labels, dtype=torch.long)

	def __getitem__(self, index: int):
		return self.tokens[index], self.labels[index]

	def __len__(self):
		return len(self.labels)


class TextDataset(Dataset):
	""" 返回的是(text, label) 不是Tensor, 必须经过TextVecCollator """

	def __init__(self, texts: Union[List[str], np.ndarray, pd.Series], labels: Union[List, np.ndarray, pd.Series]):
		super().__init__()
		self.texts = texts.values if isinstance(texts, pd.Series) else texts
		self.labels = labels.values if isinstance(labels, pd.Series) else labels

	def __getitem__(self, index: int):
		return self.texts[index], self.labels[index]

	def __len__(self):
		return len(self.labels)


class TextDFDataset(Dataset):
	""" 返回的是(text, label) 不是Tensor, 必须经过TextVecCollator """

	def __init__(self, data_df: pd.DataFrame):
		super().__init__()
		self.data = data_df.values

	def __getitem__(self, index: int):
		return self.data[index]

	def __len__(self):
		return len(self.data)


class TextVecCollator:

	def __init__(self, tokenize_vec: Union[TokenizeVec, TokenEmbedding], max_length: int = None, **kwargs):
		self.tokenize_vec = tokenize_vec
		self.max_length = max_length
		self.kwargs = kwargs

	def __call__(self, examples):
		texts, labels = zip(*examples)
		labels = torch.tensor(labels, dtype=torch.long)

		if isinstance(self.tokenize_vec, TokenizeVec):
			max_length = get_texts_max_length(texts, cut_type='char') + 2
			if self.max_length is not None:
				max_length = min(max_length, self.max_length)
			return self.tokenize_vec.encode_plus(texts, max_length=max_length, padding='max_length',
												truncation=True, add_special_tokens=True,
												return_token_type_ids=True,return_attention_mask=True,
												return_tensors='pt', **self.kwargs), labels
		elif isinstance(self.tokenize_vec, TokenEmbedding):
			max_length = get_texts_max_length(texts, cut_type=self.tokenize_vec.cut_type,
											  language=self.tokenize_vec.language, cut_fn=self.tokenize_vec.cut_fn)
			if self.max_length is not None:
				max_length = min(max_length, self.max_length)
			return self.tokenize_vec(texts, max_length, **self.kwargs), labels

		raise ValueError("Invalid tokenize_vec, it must be a TokenizeVec or TokenEmbedding.")


class TokenizeCollator:

	def __init__(self, tokenizer, max_length: int = None, **kwargs):
		self.tokenizer = tokenizer
		self.max_length = max_length
		self.kwargs = kwargs

	def __call__(self, examples):
		texts, labels = zip(*examples)
		labels = torch.tensor(labels, dtype=torch.long)

		if isinstance(self.tokenizer, BaseTokenizer):
			max_length = get_texts_max_length(texts, cut_type=self.tokenizer.cut_type, language=self.tokenizer.language,
											  cut_fn=self.tokenizer.cut_fn)
			if self.max_length is not None:
				max_length = min(max_length, self.max_length)
			return torch.tensor(self.tokenizer.batch_encode(texts, max_length, **self.kwargs), dtype=torch.long), labels

		max_length = get_texts_max_length(texts, cut_type='char') + 2
		if self.max_length is not None:
			max_length = min(max_length, self.max_length)
		result = self.tokenizer.batch_encode_plus(texts, max_length=max_length, padding='max_length',
												  return_token_type_ids=True, return_attention_mask=True,
											  	  truncation=True, add_special_tokens=True, return_tensors='pt', **self.kwargs)
		result['labels'] = labels
		return result


class PaddingTokenCollator:
	"""
	与TokenDataset配合使用
	"""

	def __init__(self, padding_func, max_length: int = None):
		self.padding_func = padding_func
		self.max_length = max_length

	def __call__(self, examples):
		tokens, labels = zip(*examples)
		labels = torch.tensor(labels, dtype=torch.long)

		max_length = max(map(lambda x: len(x), tokens))
		if self.max_length is not None:
			max_length = min(max_length, self.max_length)

		return torch.tensor(self.padding_func(tokens, max_length), dtype=torch.long), labels
